#
# python_grabber
#
# Authors:
#  Andrea Schiavinato <andrea.schiavinato84@gmail.com>
#
# Copyright (C) 2019 Andrea Schiavinato
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

from comtypes import GUID
from comtypes import IPersist
from comtypes import IUnknown
from comtypes import COMMETHOD, HRESULT, POINTER, c_int, c_ulong
from ctypes.wintypes import _ULARGE_INTEGER


class IPersistStream(IPersist):
    _case_insensitive_ = True
    _iid_ = GUID("{00000109-0000-0000-C000-000000000046}")
    _idlflags_ = []


class ISequentialStream(IUnknown):
    _case_insensitive_ = True
    _iid_ = GUID("{0C733A30-2A1C-11CE-ADE5-00AA0044773D}")
    _idlflags_ = []


class IStream(ISequentialStream):
    _case_insensitive_ = True
    _iid_ = GUID("{0000000C-0000-0000-C000-000000000046}")
    _idlflags_ = []


IPersistStream._methods_ = [
    COMMETHOD([], HRESULT, "IsDirty"),
    COMMETHOD([], HRESULT, "Load", (["in"], POINTER(IStream), "pstm")),
    COMMETHOD(
        [],
        HRESULT,
        "Save",
        (["in"], POINTER(IStream), "pstm"),
        (["in"], c_int, "fClearDirty"),
    ),
    COMMETHOD(
        [], HRESULT, "GetSizeMax", (["out"], POINTER(_ULARGE_INTEGER), "pcbSize")
    ),
]


class IEnumMoniker(IUnknown):
    _case_insensitive_ = True
    _iid_ = GUID("{00000102-0000-0000-C000-000000000046}")
    _idlflags_ = []


class IMoniker(IPersistStream):
    _case_insensitive_ = True
    _iid_ = GUID("{0000000F-0000-0000-C000-000000000046}")
    _idlflags_ = []


class IBindCtx(IUnknown):
    _case_insensitive_ = True
    _iid_ = GUID("{0000000E-0000-0000-C000-000000000046}")
    _idlflags_ = []


IEnumMoniker._methods_ = [
    COMMETHOD(
        [],
        HRESULT,
        "Next",
        (["in"], c_ulong, "celt"),
        (["out"], POINTER(POINTER(IMoniker)), "rgelt"),
        (["out"], POINTER(c_ulong), "pceltFetched"),
    ),
    COMMETHOD([], HRESULT, "Skip", (["in"], c_ulong, "celt")),
    COMMETHOD([], HRESULT, "Reset"),
    COMMETHOD([], HRESULT, "Clone", (["out"], POINTER(POINTER(IMoniker)), "ppenum")),
]

IMoniker._methods_ = [
    COMMETHOD(
        [],
        HRESULT,
        "BindToObject",
        (["in"], POINTER(IBindCtx), "pbc"),
        (["in"], POINTER(IMoniker), "pmkToLeft"),
        (["in"], POINTER(GUID), "riidResult"),
        (["out"], POINTER(POINTER(IUnknown)), "ppvResult"),
    ),
    COMMETHOD(
        [],
        HRESULT,
        "BindToStorage",
        (["in"], POINTER(IBindCtx), "pbc"),
        (["in"], POINTER(IMoniker), "pmkToLeft"),
        (["in"], POINTER(GUID), "riid"),
        (["out"], POINTER(POINTER(IUnknown)), "ppvObj"),
    ),
]

# IMoniker._methods_ = IPersistStream._methods_ + [
#    STDMETHOD(HRESULT, "BindToObject", POINTER(IBindCtx), POINTER(IMoniker), REFIID, c_void_p),
#    STDMETHOD(HRESULT, "BindToStorage", POINTER(IBindCtx), POINTER(IMoniker), REFIID, c_void_p),
#    STDMETHOD(HRESULT, "Reduce", POINTER(IBindCtx), DWORD,
#              POINTER(POINTER(IMoniker)), POINTER(POINTER(IMoniker))),
#    STDMETHOD(HRESULT, "ComposeWith", POINTER(IMoniker), BOOL, POINTER(POINTER(IMoniker))),
#    STDMETHOD(HRESULT, "Enum", BOOL, POINTER(IEnumMoniker)),
#    STDMETHOD(HRESULT, "IsEqual", POINTER(IMoniker)),
#    STDMETHOD(HRESULT, "Hash", POINTER(DWORD)),
#    STDMETHOD(HRESULT, "IsRunning", POINTER(IBindCtx), POINTER(IMoniker), POINTER(IMoniker)),
#    STDMETHOD(HRESULT, "GetTimeOfLastChange", POINTER(IBindCtx), POINTER(IMoniker),
#              POINTER(FILETIME)),
#    STDMETHOD(HRESULT, "Inverse", POINTER(IMoniker)),
#    STDMETHOD(HRESULT, "CommonPrefixWith", POINTER(IMoniker), POINTER(POINTER(IMoniker))),
#    STDMETHOD(HRESULT, "RelativePathTo", POINTER(IMoniker), POINTER(POINTER(IMoniker))),
#    STDMETHOD(HRESULT, "GetDisplayName", POINTER(IBindCtx), POINTER(IMoniker), POINTER(LPOLESTR)),
#    STDMETHOD(HRESULT, "ParseDisplayName", POINTER(IBindCtx), POINTER(IMoniker),
#              LPOLESTR, POINTER(ULONG), POINTER(POINTER(IMoniker))),
#    STDMETHOD([], HRESULT, "IsSystemMoniker", POINTER(DWORD))]



#
# python_grabber
#
# Authors:
#  Andrea Schiavinato <andrea.schiavinato84@gmail.com>
#
# Copyright (C) 2019 Andrea Schiavinato
#
# Permission is hereby granted, free of charge, to any person obtaining
# a copy of this software and associated documentation files (the
# "Software"), to deal in the Software without restriction, including
# without limitation the rights to use, copy, modify, merge, publish,
# distribute, sublicense, and/or sell copies of the Software, and to
# permit persons to whom the Software is furnished to do so, subject to
# the following conditions:
#
# The above copyright notice and this permission notice shall be
# included in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
# EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
# MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
# NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
# LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
# OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
# WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
#

from qgrabber.pygrabber.win_common_types import *
from comtypes import DWORD
from ctypes.wintypes import BOOL


LPCWSTR_WMSDK_TYPE_SAFE = POINTER(c_wchar)


class IWMStreamList(IUnknown):
    _case_insensitive_ = True
    _iid_ = GUID("{96406Bdd-2b2b-11d3-b36b-00c04f6108ff}")
    _idlflags_ = []


IWMStreamList._methods_ = [
    COMMETHOD(
        [],
        HRESULT,
        "GetStreams",
        (["in"], POINTER(DWORD), "pwStreamNumArray"),
        (["in"], POINTER(DWORD), "pcStreams"),
    ),
    COMMETHOD([], HRESULT, "AddStream", (["in"], DWORD, "wStreamNum")),
    COMMETHOD([], HRESULT, "RemoveStream", (["in"], DWORD, "wStreamNum")),
]


class IWMMutualExclusion(IWMStreamList):
    _case_insensitive_ = True
    _iid_ = GUID("{96406Bde-2b2b-11d3-b36b-00c04f6108ff}")
    _idlflags_ = []


IWMMutualExclusion._methods_ = [
    COMMETHOD([], HRESULT, "GetType", (["out"], POINTER(GUID), "pguidType")),
    COMMETHOD([], HRESULT, "SetType", (["in"], REFGUID, "guidType")),
]


class IWMStreamConfig(IUnknown):
    _case_insensitive_ = True
    _iid_ = GUID("{96406Bdc-2b2b-11d3-b36b-00c04f6108ff}")
    _idlflags_ = []


IWMStreamConfig._methods_ = [
    COMMETHOD(
        [], HRESULT, "GetStreamType", (["out"], POINTER(GUID), "pguidStreamType")
    ),
    COMMETHOD([], HRESULT, "GetStreamNumber", (["out"], POINTER(WORD), "pwStreamNum")),
    COMMETHOD([], HRESULT, "SetStreamNumber", (["in"], WORD, "wStreamNum")),
    COMMETHOD(
        [],
        HRESULT,
        "GetStreamName",
        (["in"], POINTER(c_wchar), "pwszStreamName"),
        (["in"], POINTER(WORD), "pcchStreamName"),
    ),
    COMMETHOD(
        [],
        HRESULT,
        "SetStreamName",
        (["in"], LPCWSTR_WMSDK_TYPE_SAFE, "pwszStreamName"),
    ),
    COMMETHOD(
        [],
        HRESULT,
        "GetConnectionName",
        (["out"], POINTER(c_wchar), "pwszInputName"),
        (["in"], POINTER(WORD), "pcchInputName"),
    ),
    COMMETHOD(
        [],
        HRESULT,
        "SetConnectionName",
        (["in"], LPCWSTR_WMSDK_TYPE_SAFE, "pwszInputName"),
    ),
    COMMETHOD([], HRESULT, "GetBitrate", (["out"], POINTER(WORD), "pdwBitrate")),
    COMMETHOD([], HRESULT, "SetBitrate", (["in"], DWORD, "pdwBitrate")),
    COMMETHOD(
        [], HRESULT, "pmsBufferWindow", (["out"], POINTER(WORD), "pmsBufferWindow")
    ),
    COMMETHOD([], HRESULT, "SetBufferWindow", (["in"], DWORD, "msBufferWindow")),
]


class IWMProfile(IUnknown):
    _case_insensitive_ = True
    _iid_ = GUID("{96406BDB-2B2B-11D3-B36B-00C04F6108FF}")
    _idlflags_ = []


IWMProfile._methods_ = [
    COMMETHOD([], HRESULT, "GetVersion", (["out"], POINTER(DWORD), "pdwVersion")),
    COMMETHOD(
        [],
        HRESULT,
        "GetName",
        (["in"], POINTER(c_wchar), "pwszName"),
        (["in"], POINTER(DWORD), "pcchName"),
    ),
    COMMETHOD([], HRESULT, "SetName", (["in"], POINTER(c_wchar), "pwszName")),
    COMMETHOD(
        [],
        HRESULT,
        "GetDescription",
        (["in"], POINTER(c_wchar), "pwszDescription"),
        (["in"], POINTER(DWORD), "pcchDescription"),
    ),
    COMMETHOD(
        [], HRESULT, "SetDescription", (["in"], POINTER(c_wchar), "pwszDescription")
    ),
    COMMETHOD([], HRESULT, "GetStreamCount", (["out"], POINTER(DWORD), "pcStreams")),
    COMMETHOD(
        [],
        HRESULT,
        "GetStream",
        (["in"], DWORD, "dwStreamIndex"),
        (["out"], POINTER(POINTER(IWMStreamConfig)), "ppConfig"),
    ),
    COMMETHOD(
        [],
        HRESULT,
        "GetStreamByNumber",
        (["in"], DWORD, "wStreamNum"),
        (["out"], POINTER(POINTER(IWMStreamConfig)), "ppConfig"),
    ),
    COMMETHOD(
        [], HRESULT, "RemoveStream", (["in"], POINTER(IWMStreamConfig), "pConfig")
    ),
    COMMETHOD([], HRESULT, "RemoveStreamByNumber", (["in"], DWORD, "wStreamNum")),
    COMMETHOD([], HRESULT, "AddStream", (["in"], POINTER(IWMStreamConfig), "pConfig")),
    COMMETHOD(
        [], HRESULT, "ReconfigStream", (["in"], POINTER(IWMStreamConfig), "pConfig")
    ),
    COMMETHOD(
        [],
        HRESULT,
        "CreateNewStream",
        (["in"], REFGUID, "guidStreamType"),
        (["out"], POINTER(POINTER(IWMStreamConfig)), "ppConfig"),
    ),
    COMMETHOD(
        [], HRESULT, "GetMutualExclusionCount", (["out"], POINTER(DWORD), "pcME")
    ),
    COMMETHOD(
        [],
        HRESULT,
        "GetMutualExclusion",
        (["in"], DWORD, "dwMEIndex"),
        (["out"], POINTER(POINTER(IWMMutualExclusion)), "ppME"),
    ),
    COMMETHOD(
        [],
        HRESULT,
        "RemoveMutualExclusion",
        (["in"], POINTER(IWMMutualExclusion), "pME"),
    ),
    COMMETHOD(
        [], HRESULT, "AddMutualExclusion", (["in"], POINTER(IWMMutualExclusion), "pME")
    ),
    COMMETHOD(
        [],
        HRESULT,
        "CreateNewMutualExclusion",
        (["out"], POINTER(POINTER(IWMMutualExclusion)), "ppME"),
    ),
]


class IWMProfileManager2(IUnknown):
    _case_insensitive_ = True
    _iid_ = GUID("{d16679f2-6ca0-472d-8d31-2f5d55aee155}")
    _idlflags_ = []


IWMProfileManager2._methods_ = [
    COMMETHOD(
        [],
        HRESULT,
        "CreateEmptyProfile",
        (["in"], DWORD, "dwVersion"),
        (["out"], POINTER(POINTER(IWMProfile)), "ppProfile"),
    ),
    COMMETHOD(
        [],
        HRESULT,
        "LoadProfileByID",
        (["in"], REFGUID, "guidProfile"),
        (["out"], POINTER(POINTER(IWMProfile)), "ppProfile"),
    ),
    COMMETHOD(
        [],
        HRESULT,
        "LoadProfileByData",
        (["in"], c_wchar_p, "pwszProfile"),
        (["out"], POINTER(POINTER(IWMProfile)), "ppProfile"),
    ),
    COMMETHOD(
        [],
        HRESULT,
        "SaveProfile",
        (["in"], POINTER(IWMProfile), "pIWMProfile"),
        (["in"], c_wchar_p, "pwszProfile"),
        (["in"], POINTER(DWORD), "pdwLength"),
    ),
    COMMETHOD(
        [], HRESULT, "GetSystemProfileCount", (["out"], POINTER(DWORD), "pcProfiles")
    ),
    COMMETHOD(
        [],
        HRESULT,
        "LoadSystemProfile",
        (["in"], DWORD, "dwProfileIndex"),
        (["out"], POINTER(POINTER(IWMProfile)), "ppProfile"),
    ),
    COMMETHOD(
        [], HRESULT, "GetSystemProfileVersion", (["out"], POINTER(DWORD), "pdwVersion")
    ),
    COMMETHOD([], HRESULT, "SetSystemProfileVersion", (["in"], DWORD, "dwVersion")),
]


WMCreateProfileManager = windll.Wmvcore.WMCreateProfileManager
WMCreateProfileManager.restype = HRESULT
WMCreateProfileManager.argtypes = [POINTER(POINTER(IWMProfileManager2))]


class IConfigAsfWriter(IUnknown):
    _case_insensitive_ = True
    _iid_ = GUID("{45086030-F7E4-486a-B504-826BB5792A3B}")
    _idlflags_ = []


IConfigAsfWriter._methods_ = [
    COMMETHOD(
        [], HRESULT, "ConfigureFilterUsingProfileId", (["in"], DWORD, "dwProfileId")
    ),
    COMMETHOD(
        [], HRESULT, "GetCurrentProfileId", (["out"], POINTER(DWORD), "pdwProfileId")
    ),
    COMMETHOD(
        [], HRESULT, "ConfigureFilterUsingProfileGuid", (["in"], REFGUID, "guidProfile")
    ),
    COMMETHOD(
        [], HRESULT, "GetCurrentProfileGuid", (["out"], POINTER(GUID), "pProfileGuid")
    ),
    COMMETHOD(
        [],
        HRESULT,
        "ConfigureFilterUsingProfile",
        (["in"], POINTER(IWMProfile), "pProfile"),
    ),
    COMMETHOD(
        [],
        HRESULT,
        "GetCurrentProfile",
        (["out"], POINTER(POINTER(IWMProfile)), "ppProfile"),
    ),
    COMMETHOD([], HRESULT, "SetIndexMode", (["in"], BOOL, "bIndexFile")),
    COMMETHOD([], HRESULT, "GetIndexMode", (["out"], POINTER(BOOL), "pbIndexFile")),
]



import logging

from qgrabber.pygrabber.dshow_graph import SampleGrabberCallback
import numpy as np

_LOGGER = logging.getLogger(__name__)


class WxGrabber(SampleGrabberCallback):
    def BufferCB(self, this, SampleTime, pBuffer, BufferLen):
        img = np.ctypeslib.as_array(
            pBuffer, shape=(self.image_resolution[1], self.image_resolution[0], 3)
        )
        self.callback(img)



import logging
import wx
from PIL import Image

_LOGGER = logging.getLogger(__name__)


class ScannerView(wx.Panel):
    """Panel on which the webcam data is painted."""

    def __init__(
        self,
        parent,
        controller,
        mirror_x=True,
        mirror_y=True,
        width=800,
        height=600,
        style=wx.NO_BORDER,
    ):
        """

        Args:
            parent: wx parent.
            mirror_x: Flip image on its x axis.
            mirror_y: Flip image on its y axis.
            width: width of this panel.
            height: height of this panel.
            style: a style.
        """

        wx.Panel.__init__(self, parent, size=(width, height), style=style)

        self.width = width
        self.height = height

        self.buffer = wx.NullBitmap

        self._mirror_x = mirror_x
        self._mirror_y = mirror_y

        self._controller = controller
        self._controller.frame_data.subscribe(self.set_frame)

    def on_show(self, event):
        if event.IsShown():
            self.GetParent().Layout()
            self.Layout()

    def set_frame(self, frame):
        """Populate the image with raw data."""

        if self._mirror_x:
            scan_dir = -1
        else:
            scan_dir = 1

        im = Image.frombytes(
            "RGB", (self.width, self.height), frame, "raw", "BGR", 0, scan_dir,
        )
        if self._mirror_y:
            im = im.transpose(Image.FLIP_LEFT_RIGHT)

        frame = im.tobytes()

        self.buffer = wx.Bitmap.FromBuffer(self.width, self.height, frame)
        dc = wx.BufferedDC(wx.ClientDC(self), wx.NullBitmap, wx.BUFFER_VIRTUAL_AREA)
        dc.Clear()
        dc.DrawBitmap(self.buffer, 0, 0)



import logging
from typing import List

from qgrabber.pygrabber.dshow_graph import FilterGraph, FilterType
from qgrabber.pygrabber.dshow_ids import MediaTypes, MediaSubtypes
from pyzbar.pyzbar import Decoded
from pyzbar.pyzbar import decode
from qgrabber.grabber import WxGrabber

_LOGGER = logging.getLogger(__name__)


def cropper(frame, width, height):
    if width is None:
        return frame
    y, x = frame.shape[:2]
    startx = x // 2 - width // 2
    starty = y // 2 - height // 2
    return frame[starty : starty + height, startx : startx + width].copy(order="C")


class ScannerModel:
    def __init__(self, stop_on_scan=False):
        self._graph = None
        self._crop_x = None
        self._crop_y = None
        self._camera_callback = None
        self._stop_on_scan = stop_on_scan
        self.on_code_scanned = []

    def start(self, camera_callback, on_code_scanned, crop_x=None, crop_y=None):
        """Start the scanner.

        Args:
            camera_callback: Called whenever image data from the camera is received.
            on_code_scanned: Called when a code is detected.
            crop_x: image data crop_x
            crop_y: image data crop_y

        Returns:

        """
        self._camera_callback = camera_callback
        self._crop_x = crop_x
        self._crop_y = crop_y
        self.on_code_scanned.append(on_code_scanned)

        self._graph = FilterGraph()
        self._graph.add_video_input_device(0)
        filter_type = FilterType.sample_grabber

        _filter = self._graph.filter_factory.build_filter(filter_type, None)
        self._graph.filters[filter_type] = _filter
        self._graph.filter_graph.AddFilter(_filter.instance, _filter.Name)

        sample_grabber = self._graph.filters[FilterType.sample_grabber]
        sample_grabber_cb = WxGrabber(self.call_back)

        sample_grabber.set_callback(sample_grabber_cb, 1)
        sample_grabber.set_media_type(MediaTypes.Video, MediaSubtypes.RGB24)

        self._graph.add_null_render()
        self._graph.prepare_preview_graph()
        self._graph.run()

    def stop(self, result: List["Decoded"]):
        """Stop the webcam."""
        for scanned_callback in self.on_code_scanned:
            scanned_callback(result)

    def decode(self, frame):
        codes = decode(frame)
        if codes:
            _LOGGER.debug(codes)
            self.stop(codes)
            if self._stop_on_scan:
                _LOGGER.debug("Stopping scanner.")
                self._graph.stop()

    def call_back(self, frame):
        frame = cropper(frame, self._crop_x, self._crop_y)
        self.decode(frame)
        self._camera_callback(frame)



import logging
from typing import TYPE_CHECKING, Optional

from aiosubpub import Channel

if TYPE_CHECKING:
    from pyzbar.pyzbar import Decoded
    from qgrabber.model import ScannerModel
    from qgrabber.view import ScannerView

_LOGGER = logging.getLogger(__name__)


class Controller:
    def __init__(self, model: "ScannerModel", view_width, view_height):
        self._model = model
        self.frame_data = Channel("Webcam view data")
        self._view_width = view_width
        self._view_height = view_height

    def start_scan(self):
        self._model.start(
            self._incoming_frame_data,
            on_code_scanned=self.stop_scan,
            crop_x=self._view_width,
            crop_y=self._view_height,
        )

    def _incoming_frame_data(self, frame_data):
        self.frame_data.publish(frame_data)

    def stop_scan(self, result: Optional["Decoded"] = None):
        if result:
            print(result)



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['qgrabber', 'qgrabber.pygrabber']

package_data = \
{'': ['*']}

install_requires = \
['aio-wx-widgets>=0.0.4', 'comtypes>=1.1.7', 'pyzbar', 'numpy', 'aiosubpub']

extras_require = \
{'dev': ['pylint', 'nox', 'pytest-cov', 'coverage[toml]>=5.0.3', 'pre-commit'],
 'test': ['pytest', 'pytest-asyncio', 'pytest-mock']}

setup(name='qgrabber',
      version='0.0.1',
      description='Scan QR codes using your webcam, pyzbar and wxpython. No usage of opencv',
      author='Sander Teunissen',
      author_email='no_mail@nomail.com',
      url='https://github.com/sander76/qrabber',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      extras_require=extras_require,
      python_requires='>=3.7',
     )
